@@ -7,7 +7,7 @@ use bytes::Bytes;
 use destream::{de, en};
 use futures::stream::{self, StreamExt};
 use futures::{join, try_join, TryFutureExt, TryStreamExt};
-use log::debug;
+use log::{debug, error};
 use safecast::*;
 
 use tc_btree::BTreeInstance;
@@ -72,10 +72,14 @@ impl History {
         let txn_id = *txn.id();
         let value = self.save_state(txn, value).await?;
 
-        debug!("History::append_put {} {} {} {}", txn_id, path, key, value);
+        debug!(
+            "History::append_put {} {} {:?} {:?}",
+            txn_id, path, key, value
+        );
 
         let mut block = self.write_latest(txn_id).await?;
         block.append_put(txn_id, path, key, value);
+
         Ok(())
     }
 
@@ -95,13 +99,16 @@ impl History {
                     let schema = btree.schema().to_vec();
                     let classpath = BTreeType::default().path();
 
-                    if !self.dir.contains(&txn_id, &hash).await? {
+                    if self.dir.contains(&txn_id, &hash).await? {
+                        debug!("BTree with hash {} is already saved", hash);
+                    } else {
                         let file = self
                             .dir
                             .create_file(txn_id, hash.clone(), btree.class())
                             .await?;
 
                         BTreeFile::copy_from(btree, file, txn).await?;
+                        debug!("saved BTree with hash {}", hash);
                     }
 
                     Ok(OpRef::Get((
@@ -115,9 +122,12 @@ impl History {
                     let schema = table.schema().clone();
                     let classpath = TableType::default().path();
 
-                    if !self.dir.contains(&txn_id, &hash).await? {
+                    if self.dir.contains(&txn_id, &hash).await? {
+                        debug!("Table with hash {} is already saved", hash);
+                    } else {
                         let dir = self.dir.create_dir(txn_id, hash.clone()).await?;
                         TableIndex::copy_from(table, dir, txn).await?;
+                        debug!("saved Table with hash {}", hash);
                     }
 
                     Ok(OpRef::Get((
@@ -128,32 +138,47 @@ impl History {
                 }
 
                 #[cfg(feature = "tensor")]
-                Collection::Tensor(Tensor::Dense(tensor)) => {
-                    let hash = tensor.hash_hex(&txn).await?.parse()?;
-                    let shape: Tuple<Value> = tensor
-                        .shape()
-                        .to_vec()
-                        .into_iter()
-                        .map(Value::from)
-                        .collect();
+                Collection::Tensor(tensor) => {
+                    let shape = tensor.shape().clone();
+                    let dtype = tensor.dtype();
+                    let schema = tc_tensor::Schema { shape, dtype };
+                    let classpath = tensor.class().path();
 
-                    let dtype = tc_value::ValueType::from(tensor.dtype()).path();
-                    let classpath = TensorType::Dense.path();
+                    let hash = match tensor {
+                        Tensor::Dense(dense) => {
+                            let hash = dense.hash_hex(&txn).await?.parse()?;
 
-                    if !self.dir.contains(&txn_id, &hash).await? {
-                        let file = self
-                            .dir
-                            .create_file(txn_id, hash.clone(), TensorType::Dense)
-                            .await?;
+                            if self.dir.contains(&txn_id, &hash).await? {
+                                debug!("Tensor with hash {} is already saved", hash);
+                            } else {
+                                let file = self
+                                    .dir
+                                    .create_file(txn_id, hash.clone(), TensorType::Dense)
+                                    .await?;
 
-                        DenseTensor::copy_from(tensor, file, txn).await?;
-                    }
+                                DenseTensor::copy_from(dense, file, txn).await?;
+                                debug!("saved Tensor with hash {}", hash);
+                            }
 
-                    Ok(OpRef::Get((
-                        (hash.into(), classpath).into(),
-                        Value::Tuple(vec![shape.into(), dtype.into()].into()).into(),
-                    ))
-                    .into())
+                            hash
+                        }
+                        Tensor::Sparse(sparse) => {
+                            let hash = sparse.hash_hex(&txn).await?.parse()?;
+
+                            if self.dir.contains(&txn_id, &hash).await? {
+                                debug!("Tensor with hash {} is already saved", hash);
+                            } else {
+                                let dir = self.dir.create_dir(txn_id, hash.clone()).await?;
+                                SparseTensor::copy_from(sparse, dir, txn).await?;
+                                debug!("saved Tensor with hash {}", hash);
+                            }
+
+                            hash
+                        }
+                    };
+
+                    let schema: Value = schema.cast_into();
+                    Ok(OpRef::Get(((hash.into(), classpath).into(), schema.into())).into())
                 }
             },
             State::Scalar(value) => Ok(value),
@@ -295,22 +320,14 @@ impl History {
                             subject.delete(txn, &path, key).await
                         }
                         Mutation::Put(path, key, value) => {
-                            other
-                                .resolve(txn, value)
-                                .and_then(|state| self.save_state(txn.clone(), state))
-                                .and_then(|value| {
-                                    if append {
-                                        dest.append_put(
-                                            *past_txn_id,
-                                            path.clone(),
-                                            key.clone(),
-                                            value.clone(),
-                                        );
-                                    }
+                            let value = other.resolve(txn, value).await?;
+                            let value_ref = self.save_state(txn.clone(), value.clone()).await?;
 
-                                    subject.put(txn, &path, key, value.into())
-                                })
-                                .await
+                            if append {
+                                dest.append_put(*past_txn_id, path.clone(), key.clone(), value_ref);
+                            }
+
+                            subject.put(txn, &path, key, value).await
                         }
                     };
 
@@ -325,6 +342,9 @@ impl History {
 
             let (source_hash, dest_hash) = try_join!(source.hash(), dest.hash())?;
             if source_hash != dest_hash {
+                debug!("source {:?}", &*source);
+                debug!("dest {:?}", &*dest);
+
                 return Err(TCError::bad_request(
                     "error replicating chain",
                     format!("hashes diverge at block {}", i),
@@ -332,6 +352,7 @@ impl History {
             }
 
             i += 1;
+
             if other.contains_block(txn.id(), i).await? {
                 self.create_next_block(*txn.id()).await?;
             } else {
@@ -343,6 +364,8 @@ impl History {
     }
 
     pub async fn resolve(&self, txn: &Txn, scalar: Scalar) -> TCResult<State> {
+        debug!("History::resolve {}", scalar);
+
         type OpSubject = crate::scalar::Subject;
 
         if let Scalar::Ref(tc_ref) = scalar {
@@ -355,18 +378,15 @@ impl History {
                     .map_ok(State::from)
                     .await
             } else {
+                error!("invalid subject for historical Chain state {}", tc_ref);
+
                 Err(TCError::internal(format!(
                     "invalid subject for historical Chain state {}",
                     tc_ref
                 )))
             }
-        } else if !scalar.is_ref() {
-            Ok(scalar.into())
         } else {
-            Err(TCError::internal(format!(
-                "invalid subject for historical Chain state {}",
-                scalar
-            )))
+            Ok(scalar.into())
         }
     }
 
@@ -377,6 +397,8 @@ impl History {
         schema: Scalar,
         class: CollectionType,
     ) -> TCResult<Collection> {
+        debug!("resolve historical collection value of type {}", class);
+
         match class {
             CollectionType::BTree(_) => {
                 fn schema_err<I: fmt::Display>(info: I) -> TCError {
@@ -407,19 +429,47 @@ impl History {
                 let schema = Value::try_cast_from(schema, |v| schema_err(v))?;
                 let schema = schema.try_cast_into(|v| schema_err(v))?;
 
-                let dir = self.dir.get_dir(txn.id(), &hash).await?.ok_or_else(|| {
+                let dir = self.dir.get_dir(txn.id(), &hash).await?;
+                let dir = dir.ok_or_else(|| {
                     TCError::internal(format!("missing historical Chain state {}", hash))
                 })?;
 
+                debug!("dir contents {:?}", dir.entry_ids(txn.id()).await?);
                 let table = TableIndex::load(txn, schema, dir).await?;
                 Ok(Collection::Table(table.into()))
             }
 
             #[cfg(feature = "tensor")]
-            CollectionType::Tensor(tt) => Err(TCError::not_implemented(format!(
-                "History::resolve instance of {}",
-                tt
-            ))),
+            CollectionType::Tensor(tt) => {
+                let schema: Value = schema.try_cast_into(|s| {
+                    TCError::internal(format!("invalid Tensor schema: {}", s))
+                })?;
+                let schema = schema.try_cast_into(|v| {
+                    TCError::internal(format!("invalid Tensor schema: {}", v))
+                })?;
+
+                match tt {
+                    TensorType::Dense => {
+                        let file = self.dir.get_file(txn.id(), &hash).await?;
+                        let file = file.ok_or_else(|| {
+                            TCError::internal(format!("missing historical Chain state {}", hash))
+                        })?;
+
+                        let tensor = DenseTensor::load(txn, schema, file).await?;
+                        Ok(Collection::Tensor(tensor.into()))
+                    }
+                    TensorType::Sparse => {
+                        let dir = self.dir.get_dir(txn.id(), &hash).await?;
+                        let dir = dir.ok_or_else(|| {
+                            TCError::internal(format!("missing historical Chain state {}", hash))
+                        })?;
+
+                        debug!("dir contents {:?}", dir.entry_ids(txn.id()).await?);
+                        let tensor = SparseTensor::load(txn, schema, dir).await?;
+                        Ok(Collection::Tensor(tensor.into()))
+                    }
+                }
+            }
         }
     }
 }
@@ -636,6 +686,8 @@ impl<'en> IntoView<'en, fs::Dir> for History {
     type View = HistoryView<'en>;
 
     async fn into_view(self, txn: Txn) -> TCResult<Self::View> {
+        debug!("History::into_view");
+
         let txn_id = *txn.id();
         let latest = self.latest.read(&txn_id).await?;
 
@@ -644,33 +696,19 @@ impl<'en> IntoView<'en, fs::Dir> for History {
 
         let seq = stream::iter(0..((*latest) + 1))
             .map(BlockId::from)
+            .inspect(|id| debug!("encoding history block {}", id))
             .then(read_block)
             .map_ok(move |block| {
+                let this = self.clone();
                 let txn = txn.clone();
                 let map =
                     stream::iter(block.mutations().clone()).map(move |(past_txn_id, mutations)| {
-                        let txn = txn.clone();
-                        let mutations = stream::iter(mutations).then(move |op| {
-                            let txn = txn.clone();
-                            Box::pin(async move {
-                                match op {
-                                    Mutation::Delete(path, key) => {
-                                        Ok(MutationView::Delete(path, key))
-                                    }
-                                    Mutation::Put(_path, _key, value) if value.is_ref() => {
-                                        Err(TCError::not_implemented(
-                                            "resolve reference in Mutation::Put",
-                                        ))
-                                    }
-                                    Mutation::Put(path, key, value) => {
-                                        let value =
-                                            State::from(value).into_view(txn.clone()).await?;
+                        debug!("reading block mutations");
 
-                                        Ok(MutationView::Put(path, key, value))
-                                    }
-                                }
-                            })
-                        });
+                        let this = this.clone();
+                        let txn = txn.clone();
+                        let mutations = stream::iter(mutations)
+                            .then(move |op| Box::pin(load_history(this.clone(), op, txn.clone())));
 
                         let mutations: TCTryStream<'en, MutationView<'en>> = Box::pin(mutations);
                         let mutations = en::SeqStream::from(mutations);
@@ -686,6 +724,31 @@ impl<'en> IntoView<'en, fs::Dir> for History {
     }
 }
 
+async fn load_history<'a>(history: History, op: Mutation, txn: Txn) -> TCResult<MutationView<'a>> {
+    match op {
+        Mutation::Delete(path, key) => Ok(MutationView::Delete(path, key)),
+        Mutation::Put(path, key, value) if value.is_ref() => {
+            debug!("historical mutation: PUT {}: {} <- {}", path, key, value);
+
+            let value = history
+                .resolve(&txn, value)
+                .map_err(|err| {
+                    error!("unable to load historical Chain data: {}", err);
+                    err
+                })
+                .await?;
+
+            let value = value.into_view(txn).await?;
+            Ok(MutationView::Put(path, key, value))
+        }
+        Mutation::Put(path, key, value) => {
+            let value = State::from(value).into_view(txn.clone()).await?;
+
+            Ok(MutationView::Put(path, key, value))
+        }
+    }
+}
+
 type MutationViewSeq<'en> =
     en::SeqStream<TCError, MutationView<'en>, TCTryStream<'en, MutationView<'en>>>;
 
