@@ -2,13 +2,15 @@
 
 use std::convert::{TryFrom, TryInto};
 use std::fmt;
+use std::iter::FromIterator;
 
 use async_trait::async_trait;
 use destream::{de, en};
 use futures::future::TryFutureExt;
 use log::debug;
 use safecast::{TryCastFrom, TryCastInto};
-use tc_btree::BTreeType;
+
+use tc_btree::{BTreeType, Column};
 use tc_error::*;
 use tc_transact::fs::{Dir, File, Persist, Restore, Store};
 use tc_transact::{IntoView, Transact, Transaction, TxnId};
@@ -18,7 +20,9 @@ use crate::collection::{
     BTree, BTreeFile, Collection, CollectionType, Table, TableIndex, TableType,
 };
 #[cfg(feature = "tensor")]
-use crate::collection::{DenseTensor, DenseTensorFile, Tensor, TensorType};
+use crate::collection::{
+    DenseTensor, DenseTensorFile, SparseTable, SparseTensor, Tensor, TensorType,
+};
 use crate::fs;
 use crate::scalar::{Link, OpRef, Scalar, TCRef, Value, ValueType};
 use crate::state::{State, StateView};
@@ -50,7 +54,9 @@ pub enum Schema {
     BTree(tc_btree::RowSchema),
     Table(tc_table::TableSchema),
     #[cfg(feature = "tensor")]
-    Tensor(tc_tensor::Schema),
+    Dense(tc_tensor::Schema),
+    #[cfg(feature = "tensor")]
+    Sparse(tc_tensor::Schema),
     Value(Value),
 }
 
@@ -86,13 +92,18 @@ impl Schema {
                             }
 
                             #[cfg(feature = "tensor")]
-                            CollectionType::Tensor(_) => {
-                                let (shape, dtype): (Vec<u64>, ValueType) =
-                                    schema.try_cast_into(|s| {
-                                        TCError::bad_request("invalid Tensor schema", s)
-                                    })?;
+                            CollectionType::Tensor(tt) => {
+                                let schema: Value = schema.try_cast_into(|s| {
+                                    TCError::bad_request("invalid Tensor schema", s)
+                                })?;
+                                let schema = schema.try_cast_into(|v| {
+                                    TCError::bad_request("invalid Tensor schema", v)
+                                })?;
 
-                                Ok(Self::Tensor((shape.into(), dtype.try_into()?)))
+                                match tt {
+                                    TensorType::Dense => Ok(Self::Dense(schema)),
+                                    TensorType::Sparse => Ok(Self::Sparse(schema)),
+                                }
                             }
                         }
                     }
@@ -132,12 +143,9 @@ impl<'en> en::IntoStream<'en> for Schema {
                 map.end()
             }
             #[cfg(feature = "tensor")]
-            Self::Tensor((shape, dtype)) => {
+            Self::Dense(schema) | Self::Sparse(schema) => {
                 let mut map = encoder.encode_map(Some(1))?;
-                map.encode_entry(
-                    TensorType::Dense.path(),
-                    ((shape.into_vec(), ValueType::from(dtype).path()),),
-                )?;
+                map.encode_entry(TensorType::Dense.path(), (schema,))?;
                 map.end()
             }
             Self::Value(value) => value.into_stream(encoder),
@@ -145,13 +153,29 @@ impl<'en> en::IntoStream<'en> for Schema {
     }
 }
 
+impl fmt::Display for Schema {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match self {
+            Self::BTree(schema) => write!(f, "{}", Tuple::<&Column>::from_iter(schema)),
+            Self::Table(schema) => fmt::Display::fmt(schema, f),
+            Self::Value(schema) => fmt::Display::fmt(schema, f),
+            #[cfg(feature = "tensor")]
+            Self::Dense(schema) => fmt::Display::fmt(schema, f),
+            #[cfg(feature = "tensor")]
+            Self::Sparse(schema) => fmt::Display::fmt(schema, f),
+        }
+    }
+}
+
 /// The state whose transactional integrity is protected by a [`Chain`].
 #[derive(Clone)]
 pub enum Subject {
     BTree(BTreeFile),
     Table(TableIndex),
     #[cfg(feature = "tensor")]
-    Tensor(DenseTensor<DenseTensorFile>),
+    Dense(DenseTensor<DenseTensorFile>),
+    #[cfg(feature = "tensor")]
+    Sparse(SparseTensor<SparseTable>),
     Value(fs::File<Value>),
 }
 
@@ -169,20 +193,33 @@ impl Subject {
                     .await
             }
             Schema::Table(schema) => {
-                TableIndex::create(schema, dir, txn_id)
+                TableIndex::create(dir, schema, txn_id)
                     .map_ok(Self::Table)
                     .await
             }
             #[cfg(feature = "tensor")]
-            Schema::Tensor(schema) => {
+            Schema::Dense(schema) => {
                 let file = dir
                     .create_file(txn_id, SUBJECT.into(), TensorType::Dense)
                     .await?;
 
                 DenseTensor::create(file, schema, txn_id)
-                    .map_ok(Self::Tensor)
+                    .map_ok(Self::Dense)
                     .await
             }
+            #[cfg(feature = "tensor")]
+            Schema::Sparse(schema) => {
+                let dir = dir.create_dir(txn_id, SUBJECT.into()).await?;
+                let tensor = SparseTensor::create(&dir, schema, txn_id)
+                    .map_ok(Self::Sparse)
+                    .await?;
+
+                debug!(
+                    "subject dir contents after creating sparse tensor: {:?}",
+                    dir.entry_ids(&txn_id).await?
+                );
+                Ok(tensor)
+            }
             Schema::Value(value) => {
                 let file: fs::File<Value> = dir
                     .create_file(txn_id, SUBJECT.into(), value.class())
@@ -215,13 +252,25 @@ impl Subject {
                 }
             }
             #[cfg(feature = "tensor")]
-            Schema::Tensor(schema) => {
+            Schema::Dense(schema) => {
                 if let Some(file) = dir.get_file(txn.id(), &SUBJECT.into()).await? {
                     DenseTensor::load(txn, schema, file)
-                        .map_ok(Self::Tensor)
+                        .map_ok(Self::Dense)
                         .await
                 } else {
-                    Self::create(Schema::Tensor(schema), dir, *txn.id()).await
+                    Self::create(Schema::Dense(schema), dir, *txn.id()).await
+                }
+            }
+            #[cfg(feature = "tensor")]
+            Schema::Sparse(schema) => {
+                debug!("chain dir contents {:?}", dir.entry_ids(txn.id()).await?);
+                if let Some(dir) = dir.get_dir(txn.id(), &SUBJECT.into()).await? {
+                    debug!("subject dir contents {:?}", dir.entry_ids(txn.id()).await?);
+                    SparseTensor::load(txn, schema, dir)
+                        .map_ok(Self::Sparse)
+                        .await
+                } else {
+                    Self::create(Schema::Sparse(schema), dir, *txn.id()).await
                 }
             }
             Schema::Value(value) => {
@@ -250,7 +299,7 @@ impl Subject {
                 other => Err(TCError::bad_request("cannot restore a Table from", other)),
             },
             #[cfg(feature = "tensor")]
-            Self::Tensor(tensor) => match backup {
+            Self::Dense(tensor) => match backup {
                 State::Collection(Collection::Tensor(Tensor::Dense(backup))) => {
                     let file = txn
                         .context()
@@ -267,6 +316,19 @@ impl Subject {
                     other,
                 )),
             },
+            #[cfg(feature = "tensor")]
+            Self::Sparse(tensor) => match backup {
+                State::Collection(Collection::Tensor(Tensor::Sparse(backup))) => {
+                    let dir = txn.context().create_dir_tmp(txn_id).await?;
+                    let backup =
+                        tc_transact::fs::CopyFrom::copy_from(backup, dir, txn.clone()).await?;
+                    tensor.restore(&backup, txn_id).await
+                }
+                other => Err(TCError::bad_request(
+                    "cannot restore a sparse Tensor from",
+                    other,
+                )),
+            },
             Self::Value(file) => {
                 let backup = backup.try_into()?;
                 let mut block = file.write_block(txn_id, SUBJECT.into()).await?;
@@ -280,21 +342,29 @@ impl Subject {
 #[async_trait]
 impl Transact for Subject {
     async fn commit(&self, txn_id: &TxnId) {
+        debug!("commit chain subject");
+
         match self {
             Self::BTree(btree) => btree.commit(txn_id).await,
             Self::Table(table) => table.commit(txn_id).await,
             #[cfg(feature = "tensor")]
-            Self::Tensor(tensor) => tensor.commit(txn_id).await,
+            Self::Dense(tensor) => tensor.commit(txn_id).await,
+            #[cfg(feature = "tensor")]
+            Self::Sparse(tensor) => tensor.commit(txn_id).await,
             Self::Value(file) => file.commit(txn_id).await,
         }
     }
 
     async fn finalize(&self, txn_id: &TxnId) {
+        debug!("finalize chain subject");
+
         match self {
             Self::BTree(btree) => btree.finalize(txn_id).await,
             Self::Table(table) => table.finalize(txn_id).await,
             #[cfg(feature = "tensor")]
-            Self::Tensor(tensor) => tensor.finalize(txn_id).await,
+            Self::Dense(tensor) => tensor.finalize(txn_id).await,
+            #[cfg(feature = "tensor")]
+            Self::Sparse(tensor) => tensor.finalize(txn_id).await,
             Self::Value(file) => file.finalize(txn_id).await,
         }
     }
@@ -334,7 +404,9 @@ impl<'en> IntoView<'en, fs::Dir> for Subject {
             }
             Self::Table(table) => State::from(Table::Table(table)).into_view(txn).await,
             #[cfg(feature = "tensor")]
-            Self::Tensor(tensor) => State::from(Tensor::from(tensor)).into_view(txn).await,
+            Self::Dense(tensor) => State::from(Tensor::from(tensor)).into_view(txn).await,
+            #[cfg(feature = "tensor")]
+            Self::Sparse(tensor) => State::from(Tensor::from(tensor)).into_view(txn).await,
             Self::BTree(btree) => State::from(BTree::File(btree)).into_view(txn).await,
         }
     }
@@ -346,7 +418,9 @@ impl fmt::Display for Subject {
             Self::BTree(btree) => write!(f, "chain Subject, {}", btree.class()),
             Self::Table(table) => write!(f, "chain Subject, {}", table.class()),
             #[cfg(feature = "tensor")]
-            Self::Tensor(_) => write!(f, "chain Subject, {}", TensorType::Dense),
+            Self::Dense(_) => write!(f, "chain Subject, {}", TensorType::Dense),
+            #[cfg(feature = "tensor")]
+            Self::Sparse(_) => write!(f, "chain Subject, {}", TensorType::Sparse),
             Self::Value(_) => write!(f, "chain Subject, {}", ValueType::Value),
         }
     }
@@ -411,6 +485,12 @@ impl NativeClass for ChainType {
     }
 }
 
+impl fmt::Debug for ChainType {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        fmt::Display::fmt(self, f)
+    }
+}
+
 impl fmt::Display for ChainType {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         f.write_str(match self {
@@ -482,8 +562,6 @@ impl ChainInstance for Chain {
     }
 
     async fn write_ahead(&self, txn_id: &TxnId) {
-        debug!("Chain::write_ahead {}", txn_id);
-
         match self {
             Self::Block(chain) => chain.write_ahead(txn_id).await,
             Self::Sync(chain) => chain.write_ahead(txn_id).await,
@@ -540,6 +618,12 @@ impl<'en> IntoView<'en, fs::Dir> for Chain {
     }
 }
 
+impl fmt::Debug for Chain {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        write!(f, "instance of {}", self.class())
+    }
+}
+
 impl fmt::Display for Chain {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         write!(f, "instance of {}", self.class())
