@@ -26,7 +26,7 @@ pub use tc_btree::BTreeType;
 pub use tc_table::TableType;
 
 #[cfg(feature = "tensor")]
-pub use tc_tensor::{DenseAccess, TensorType};
+pub use tc_tensor::{DenseAccess, SparseAccess, TensorType};
 
 pub type BTree = tc_btree::BTree<fs::File<tc_btree::Node>, fs::Dir, Txn>;
 pub type BTreeFile = tc_btree::BTreeFile<fs::File<tc_btree::Node>, fs::Dir, Txn>;
@@ -35,11 +35,20 @@ pub type Table = tc_table::Table<fs::File<tc_btree::Node>, fs::Dir, Txn>;
 pub type TableIndex = tc_table::TableIndex<fs::File<tc_btree::Node>, fs::Dir, Txn>;
 
 #[cfg(feature = "tensor")]
-pub type Tensor = tc_tensor::Tensor<fs::File<afarray::Array>, fs::Dir, Txn>;
+pub type Tensor =
+    tc_tensor::Tensor<fs::File<afarray::Array>, fs::File<tc_btree::Node>, fs::Dir, Txn>;
 #[cfg(feature = "tensor")]
-pub type DenseTensor<B> = tc_tensor::DenseTensor<fs::File<afarray::Array>, fs::Dir, Txn, B>;
+pub type DenseTensor<B> =
+    tc_tensor::DenseTensor<fs::File<afarray::Array>, fs::File<tc_btree::Node>, fs::Dir, Txn, B>;
 #[cfg(feature = "tensor")]
-pub type DenseTensorFile = tc_tensor::BlockListFile<fs::File<afarray::Array>, fs::Dir, Txn>;
+pub type DenseTensorFile =
+    tc_tensor::BlockListFile<fs::File<afarray::Array>, fs::File<tc_btree::Node>, fs::Dir, Txn>;
+#[cfg(feature = "tensor")]
+pub type SparseTensor<A> =
+    tc_tensor::SparseTensor<fs::File<afarray::Array>, fs::File<tc_btree::Node>, fs::Dir, Txn, A>;
+#[cfg(feature = "tensor")]
+pub type SparseTable =
+    tc_tensor::SparseTable<fs::File<afarray::Array>, fs::File<tc_btree::Node>, fs::Dir, Txn>;
 
 pub const PREFIX: PathLabel = path_label(&["state", "collection"]);
 
@@ -165,12 +174,23 @@ impl From<Tensor> for Collection {
 }
 
 #[cfg(feature = "tensor")]
-impl<B: DenseAccess<fs::File<afarray::Array>, fs::Dir, Txn>> From<DenseTensor<B>> for Collection {
+impl<B: DenseAccess<fs::File<afarray::Array>, fs::File<tc_btree::Node>, fs::Dir, Txn>>
+    From<DenseTensor<B>> for Collection
+{
     fn from(tensor: DenseTensor<B>) -> Self {
         Self::Tensor(tensor.into())
     }
 }
 
+#[cfg(feature = "tensor")]
+impl<A: SparseAccess<fs::File<afarray::Array>, fs::File<tc_btree::Node>, fs::Dir, Txn>>
+    From<SparseTensor<A>> for Collection
+{
+    fn from(tensor: SparseTensor<A>) -> Self {
+        Self::Tensor(tensor.into())
+    }
+}
+
 #[cfg(feature = "tensor")]
 impl safecast::TryCastFrom<Collection> for Tensor {
     fn can_cast_from(collection: &Collection) -> bool {
@@ -225,11 +245,13 @@ impl CollectionVisitor {
             #[cfg(feature = "tensor")]
             CollectionType::Tensor(tt) => match tt {
                 TensorType::Dense => {
-                    debug!("deserialize Tensor");
-
                     let tensor: DenseTensor<DenseTensorFile> = access.next_value(self.txn).await?;
                     Ok(Collection::Tensor(tensor.into()))
                 }
+                TensorType::Sparse => {
+                    let tensor: SparseTensor<SparseTable> = access.next_value(self.txn).await?;
+                    Ok(Collection::Tensor(tensor.into()))
+                }
             },
         }
     }
@@ -280,6 +302,17 @@ impl<'en> IntoView<'en, fs::Dir> for Collection {
     }
 }
 
+impl fmt::Debug for Collection {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match self {
+            Self::BTree(btree) => fmt::Debug::fmt(btree, f),
+            Self::Table(table) => fmt::Debug::fmt(table, f),
+            #[cfg(feature = "tensor")]
+            Self::Tensor(tensor) => fmt::Debug::fmt(tensor, f),
+        }
+    }
+}
+
 impl fmt::Display for Collection {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
@@ -309,7 +342,8 @@ impl<'en> en::IntoStream<'en> for CollectionView<'en> {
             Self::Table(table) => map.encode_entry(TableType::default().path(), table),
             #[cfg(feature = "tensor")]
             Self::Tensor(tensor) => match tensor {
-                TensorView::Dense(tensor) => map.encode_entry(TensorType::Dense.path(), tensor),
+                TensorView::Dense(dense) => map.encode_entry(TensorType::Dense.path(), dense),
+                TensorView::Sparse(sparse) => map.encode_entry(TensorType::Sparse.path(), sparse),
             },
         }?;
         map.end()
