@@ -138,7 +138,44 @@ impl Kernel {
             );
 
             execute(txn, cluster, |txn, cluster| async move {
-                cluster.put(&txn, suffix, key, value).await
+                cluster
+                    .put(&txn, suffix, key.clone(), value.clone())
+                    .await?;
+
+                let txn = if !txn.has_leader(cluster.path()) {
+                    cluster.lead(txn).await?
+                } else {
+                    txn
+                };
+
+                let self_link = txn.link(cluster.path().to_vec().into());
+                if suffix.is_empty() {
+                    // it's a synchronization message
+                    return Ok(());
+                } else if !txn.is_leader(cluster.path()) {
+                    debug!(
+                        "{} successfully replicated PUT {}",
+                        self_link,
+                        TCPath::from(suffix)
+                    );
+                    return Ok(());
+                }
+
+                debug!(
+                    "{} is leading replication of PUT {}",
+                    self_link,
+                    TCPath::from(suffix)
+                );
+
+                let write = |replica_link: Link| {
+                    let mut target = replica_link.clone();
+                    target.extend(suffix.to_vec());
+
+                    debug!("replicate PUT to {}", target);
+                    txn.put(target, key.clone(), value.clone())
+                };
+
+                cluster.replicate_write(txn.clone(), write).await
             })
             .await
         } else {
@@ -221,7 +258,42 @@ impl Kernel {
             );
 
             execute(txn, cluster, |txn, cluster| async move {
-                cluster.delete(&txn, suffix, key).await
+                cluster.delete(&txn, suffix, key.clone()).await?;
+
+                let txn = if !txn.has_leader(cluster.path()) {
+                    cluster.lead(txn).await?
+                } else {
+                    txn
+                };
+
+                let self_link = txn.link(cluster.path().to_vec().into());
+                if suffix.is_empty() {
+                    // it's a synchronization message
+                    return Ok(());
+                } else if !txn.is_leader(cluster.path()) {
+                    debug!(
+                        "{} successfully replicated DELETE {}",
+                        self_link,
+                        TCPath::from(suffix)
+                    );
+
+                    return Ok(());
+                }
+
+                debug!(
+                    "{} is leading replication of DELETE {}...",
+                    self_link,
+                    TCPath::from(suffix)
+                );
+                let write = |replica_link: Link| {
+                    let mut target = replica_link.clone();
+                    target.extend(suffix.to_vec());
+
+                    debug!("replicate DELETE to {}", target);
+                    txn.delete(target, key.clone())
+                };
+
+                cluster.replicate_write(txn.clone(), write).await
             })
             .await
         } else if &path[0] == "error" && path.len() == 2 {
@@ -258,13 +330,11 @@ fn execute<
 ) -> Pin<Box<dyn Future<Output = TCResult<R>> + Send + 'a>> {
     Box::pin(async move {
         if let Some(owner) = txn.owner() {
-            if cluster.path() == &owner.path()[..] {
-                debug!(
-                    "{} owns this transaction, no need to notify",
-                    TCPath::from(cluster.path())
-                );
+            if owner.path() == cluster.path() {
+                debug!("{} owns this transaction, no need to notify", cluster);
             } else {
-                txn.put(owner.clone(), Value::None, cluster.link().clone().into())
+                let self_link = txn.link(cluster.path().to_vec().into());
+                txn.put(owner.clone(), Value::None, self_link.into())
                     .await?;
             }
 
@@ -275,8 +345,10 @@ fn execute<
             let result = handler(txn.clone(), cluster).await;
 
             if result.is_ok() {
+                debug!("commit {}", cluster);
                 cluster.distribute_commit(txn).await?;
             } else {
+                debug!("rollback {}", cluster);
                 cluster.distribute_rollback(txn).await;
             }
 
