@@ -12,7 +12,7 @@ use bytes::Bytes;
 use destream::{de, en};
 use futures::future::{try_join_all, TryFutureExt};
 use log::debug;
-use safecast::{Match, TryCastFrom, TryCastInto};
+use safecast::{TryCastFrom, TryCastInto};
 
 use tc_error::*;
 use tcgeneric::*;
@@ -257,6 +257,74 @@ impl Instance for Scalar {
 
 #[async_trait]
 impl Refer for Scalar {
+    fn dereference_self(self, path: &TCPathBuf) -> Self {
+        match self {
+            Self::Map(map) => {
+                let map = map
+                    .into_iter()
+                    .map(|(id, scalar)| (id, scalar.dereference_self(path)))
+                    .collect();
+
+                Self::Map(map)
+            }
+            Self::Op(op_def) => Self::Op(op_def.dereference_self(path)),
+            Self::Ref(tc_ref) => {
+                let tc_ref = tc_ref.dereference_self(path);
+                Self::Ref(Box::new(tc_ref))
+            }
+            Self::Tuple(tuple) => {
+                let tuple = tuple
+                    .into_iter()
+                    .map(|scalar| scalar.dereference_self(path))
+                    .collect();
+
+                Self::Tuple(tuple)
+            }
+            other => other,
+        }
+    }
+
+    fn is_inter_service_write(&self, cluster_path: &[PathSegment]) -> bool {
+        match self {
+            Self::Map(map) => map
+                .values()
+                .any(|scalar| scalar.is_inter_service_write(cluster_path)),
+            Self::Op(op_def) => op_def.is_inter_service_write(cluster_path),
+            Self::Ref(tc_ref) => tc_ref.is_inter_service_write(cluster_path),
+            Self::Tuple(tuple) => tuple
+                .iter()
+                .any(|scalar| scalar.is_inter_service_write(cluster_path)),
+            _ => false,
+        }
+    }
+
+    fn reference_self(self, path: &TCPathBuf) -> Self {
+        match self {
+            Self::Map(map) => {
+                let map = map
+                    .into_iter()
+                    .map(|(id, scalar)| (id, scalar.reference_self(path)))
+                    .collect();
+
+                Self::Map(map)
+            }
+            Self::Op(op_def) => Self::Op(op_def.reference_self(path)),
+            Self::Ref(tc_ref) => {
+                let tc_ref = tc_ref.reference_self(path);
+                Self::Ref(Box::new(tc_ref))
+            }
+            Self::Tuple(tuple) => {
+                let tuple = tuple
+                    .into_iter()
+                    .map(|scalar| scalar.reference_self(path))
+                    .collect();
+
+                Self::Tuple(tuple)
+            }
+            other => other,
+        }
+    }
+
     fn requires(&self, deps: &mut HashSet<Id>) {
         match self {
             Self::Map(map) => {
@@ -457,55 +525,17 @@ impl TryCastFrom<Scalar> for OpDef {
     }
 }
 
-impl TryCastFrom<Scalar> for Bound {
-    fn can_cast_from(scalar: &Scalar) -> bool {
-        match scalar {
-            Scalar::Map(map) if map.len() == 1 => {
-                map.contains_key(&IN.into()) || map.contains_key(&EX.into())
-            }
-            Scalar::Value(value) => value.is_none(),
-            _ => false,
-        }
-    }
-
-    fn opt_cast_from(scalar: Scalar) -> Option<Self> {
-        match scalar {
-            Scalar::Map(map) if BTreeMap::<Id, Value>::can_cast_from(&map) => {
-                let map = BTreeMap::<Id, Value>::opt_cast_from(map).unwrap();
-                Map::from(map).opt_cast_into()
-            }
-            Scalar::Value(value) if value.is_none() => Some(Self::Un),
-            _ => None,
-        }
-    }
-}
-
 impl TryCastFrom<Scalar> for Range {
     fn can_cast_from(scalar: &Scalar) -> bool {
         match scalar {
-            Scalar::Tuple(tuple) if tuple.len() == 2 => {
-                tuple.iter().all(|s| Bound::can_cast_from(s))
-            }
-            Scalar::Value(Value::Tuple(tuple)) => tuple.len() == 2,
+            Scalar::Value(value) => Self::can_cast_from(value),
             _ => false,
         }
     }
 
     fn opt_cast_from(scalar: Scalar) -> Option<Self> {
         match scalar {
-            Scalar::Tuple(tuple)
-                if tuple.len() == 2 && tuple.iter().all(|s| Bound::can_cast_from(s)) =>
-            {
-                let (start, end) = tuple.opt_cast_into().unwrap();
-                Some(Self { start, end })
-            }
-            Scalar::Value(Value::Tuple(tuple)) if tuple.len() == 2 => {
-                let (start, end) = tuple.opt_cast_into().unwrap();
-                Some(Self {
-                    start: Bound::In(start),
-                    end: Bound::Ex(end),
-                })
-            }
+            Scalar::Value(value) => value.opt_cast_into(),
             _ => None,
         }
     }
@@ -639,6 +669,24 @@ impl<T: TryCastFrom<Scalar>> TryCastFrom<Scalar> for Vec<T> {
     }
 }
 
+impl TryCastFrom<Scalar> for Tuple<Value> {
+    fn can_cast_from(scalar: &Scalar) -> bool {
+        match scalar {
+            Scalar::Tuple(tuple) => tuple.iter().all(Value::can_cast_from),
+            Scalar::Value(Value::Tuple(_)) => true,
+            _ => false,
+        }
+    }
+
+    fn opt_cast_from(scalar: Scalar) -> Option<Self> {
+        match scalar {
+            Scalar::Tuple(tuple) => tuple.into_iter().map(Value::opt_cast_from).collect(),
+            Scalar::Value(Value::Tuple(tuple)) => Some(tuple),
+            _ => None,
+        }
+    }
+}
+
 impl TryCastFrom<Scalar> for bool {
     fn can_cast_from(scalar: &Scalar) -> bool {
         match scalar {
@@ -832,30 +880,11 @@ impl ScalarVisitor {
         }
 
         let subject = Link::from(class.path()).into();
-        let op_ref = if scalar.matches::<(Scalar, Scalar)>() {
-            let (key, value) = scalar.opt_cast_into().unwrap();
-            OpRef::Put((subject, key, value))
-        } else if scalar.matches::<(Scalar,)>() {
-            let (key,) = scalar.opt_cast_into().unwrap();
-            OpRef::Get((subject, key))
-        } else if scalar.matches::<Map<Scalar>>() {
-            let params = scalar.opt_cast_into().unwrap();
-            OpRef::Post((subject, params))
-        } else {
-            return Err(de::Error::invalid_type(
-                scalar,
-                format!("an Op with subject {}", subject),
-            ));
-        };
-
-        Ok(TCRef::Op(op_ref).into())
+        Self::visit_subject(subject, scalar)
     }
 
     pub fn visit_subject<E: de::Error>(subject: Subject, params: Scalar) -> Result<Scalar, E> {
-        if let Scalar::Map(params) = params {
-            let op_ref = OpRef::Post((subject, params));
-            Ok(Scalar::Ref(Box::new(TCRef::Op(op_ref))))
-        } else if params.is_none() {
+        if params.is_none() {
             match subject {
                 Subject::Ref(id, path) if path.is_empty() => {
                     Ok(Scalar::Ref(Box::new(TCRef::Id(id))))
@@ -953,9 +982,12 @@ impl de::Visitor for ScalarVisitor {
         if key.starts_with('/') {
             if let Ok(path) = TCPathBuf::from_str(&key) {
                 if let Some(class) = ScalarType::from_path(&path) {
+                    debug!("decode instance of {}", class);
                     if let Ok(scalar) = Self::visit_map_value(class, &mut access).await {
                         return Ok(scalar);
                     }
+                } else {
+                    debug!("not a scalar classpath: {}", path);
                 }
             }
         }
@@ -1030,12 +1062,25 @@ impl<'en> en::IntoStream<'en> for Scalar {
     }
 }
 
+impl fmt::Debug for Scalar {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match self {
+            Scalar::Map(map) => fmt::Debug::fmt(map, f),
+            Scalar::Op(op) => fmt::Debug::fmt(op, f),
+            Scalar::Range(range) => fmt::Debug::fmt(range, f),
+            Scalar::Ref(tc_ref) => fmt::Debug::fmt(tc_ref, f),
+            Scalar::Tuple(tuple) => fmt::Debug::fmt(tuple, f),
+            Scalar::Value(value) => fmt::Debug::fmt(value, f),
+        }
+    }
+}
+
 impl fmt::Display for Scalar {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
             Scalar::Map(map) => fmt::Display::fmt(map, f),
             Scalar::Op(op) => fmt::Display::fmt(op, f),
-            Scalar::Range(_) => unimplemented!(),
+            Scalar::Range(range) => fmt::Display::fmt(range, f),
             Scalar::Ref(tc_ref) => fmt::Display::fmt(tc_ref, f),
             Scalar::Tuple(tuple) => fmt::Display::fmt(tuple, f),
             Scalar::Value(value) => fmt::Display::fmt(value, f),
