@@ -19,6 +19,12 @@ use crate::txn::Txn;
 
 use super::{Scalar, Scope, Value};
 
+pub use after::After;
+pub use case::Case;
+pub use id::*;
+pub use op::*;
+pub use r#if::IfRef;
+
 mod after;
 mod case;
 mod r#if;
@@ -26,17 +32,24 @@ mod r#if;
 pub mod id;
 pub mod op;
 
-pub use after::After;
-pub use case::Case;
-pub use id::*;
-pub use op::*;
-pub use r#if::IfRef;
-
 const PREFIX: PathLabel = path_label(&["state", "scalar", "ref"]);
 
 /// Trait defining dependencies and a resolution method for a [`TCRef`].
 #[async_trait]
 pub trait Refer {
+    /// Replace references to "$self" with the given relative path.
+    ///
+    /// This is used to control whether or not an OpDef will be replicated.
+    fn dereference_self(self, path: &TCPathBuf) -> Self;
+
+    /// Return `true` if this references a write operation to a cluster other than the path given.
+    fn is_inter_service_write(&self, cluster_path: &[PathSegment]) -> bool;
+
+    /// Replace the given relative path with "$self".
+    ///
+    /// This is used to control whether or not an OpDef will be replicated.
+    fn reference_self(self, path: &TCPathBuf) -> Self;
+
     /// Add the dependency [`Id`]s of this reference to the given set.
     fn requires(&self, deps: &mut HashSet<Id>);
 
@@ -127,6 +140,54 @@ impl Instance for TCRef {
 
 #[async_trait]
 impl Refer for TCRef {
+    fn dereference_self(self, path: &TCPathBuf) -> Self {
+        match self {
+            Self::After(after) => {
+                let after = after.dereference_self(path);
+                Self::After(Box::new(after))
+            }
+            Self::Case(case) => {
+                let case = case.dereference_self(path);
+                Self::Case(Box::new(case))
+            }
+            Self::Id(id_ref) => Self::Id(id_ref.dereference_self(path)),
+            Self::If(if_ref) => {
+                let if_ref = if_ref.dereference_self(path);
+                Self::If(Box::new(if_ref))
+            }
+            Self::Op(op_ref) => Self::Op(op_ref.dereference_self(path)),
+        }
+    }
+
+    fn is_inter_service_write(&self, cluster_path: &[PathSegment]) -> bool {
+        match self {
+            Self::After(after) => after.is_inter_service_write(cluster_path),
+            Self::Case(case) => case.is_inter_service_write(cluster_path),
+            Self::Id(id_ref) => id_ref.is_inter_service_write(cluster_path),
+            Self::If(if_ref) => if_ref.is_inter_service_write(cluster_path),
+            Self::Op(op_ref) => op_ref.is_inter_service_write(cluster_path),
+        }
+    }
+
+    fn reference_self(self, path: &TCPathBuf) -> Self {
+        match self {
+            Self::After(after) => {
+                let after = after.reference_self(path);
+                Self::After(Box::new(after))
+            }
+            Self::Case(case) => {
+                let case = case.reference_self(path);
+                Self::Case(Box::new(case))
+            }
+            Self::Id(id_ref) => Self::Id(id_ref.reference_self(path)),
+            Self::If(if_ref) => {
+                let if_ref = if_ref.reference_self(path);
+                Self::If(Box::new(if_ref))
+            }
+            Self::Op(op_ref) => Self::Op(op_ref.reference_self(path)),
+        }
+    }
+
     fn requires(&self, deps: &mut HashSet<Id>) {
         match self {
             Self::After(after) => after.requires(deps),
@@ -318,6 +379,18 @@ impl<'en> IntoStream<'en> for TCRef {
     }
 }
 
+impl fmt::Debug for TCRef {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match self {
+            Self::After(after) => fmt::Debug::fmt(after, f),
+            Self::Case(case) => fmt::Debug::fmt(case, f),
+            Self::Id(id_ref) => fmt::Debug::fmt(id_ref, f),
+            Self::If(if_ref) => fmt::Debug::fmt(if_ref, f),
+            Self::Op(op_ref) => fmt::Debug::fmt(op_ref, f),
+        }
+    }
+}
+
 impl fmt::Display for TCRef {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
